#!/usr/bin/env python
# coding: utf-8

# In[1]:
import numpy as np
import math
def generateWB(P,M):
    if(len(P)==1):
        print("Can't genarete an Hypersphere with one pattern")
    elif(len(P)==0):
        print("Can't genarete an Hypersphere with no patterns")
       
    center = []
   
    
    aux_point = [0]*len(P[0])

    for i in range(len(P[0])):
        Pclass_x = P[:,i] 
        vmax = max(Pclass_x.tolist())
        vmin = min(Pclass_x.tolist())
        dst = (vmax - vmin) / 2.0
        centroid = vmin + dst
        
        aux_point[i] = vmax
        
        center.append(centroid)
   
    return center

def HEpC(P,T,M=0):
    dendrite = []
    counter = 0
    classes = np.unique(T)
    
    if (len(classes)>0  and len(classes)<=2):
        covariances = np.zeros((1,np.shape(P)[1],np.shape(P)[1]))
        centroids = np.zeros((1,1,np.shape(P)[1]))
        for c in range(1):
            indexClasses = np.where(T==c)[0]
            covariances[c] = np.cov(P[indexClasses, :].T)
           

            if(len(indexClasses)==1):
                continue
            Pclass = P[indexClasses]

            center = generateWB(Pclass,M)
            

            if counter == 0:
                center_arr = center
        centroids[0][0] = np.array(center_arr)
   
      
            
    if (len(classes)>2):
        covariances = np.zeros((len(classes),np.shape(P)[1],np.shape(P)[1]))
        centroids = np.zeros((len(classes),1,np.shape(P)[1]))
        for c in classes:
            indexClasses = np.where(T==c)[0]
            covariances[c] = np.cov(P[indexClasses, :].T)
           

            if(len(indexClasses)==1):
                continue
            Pclass = P[indexClasses]

            center = generateWB(Pclass,M)
          
            centroids[counter][0] = np.array(center)
                

            counter += 1
    
    dendrite = covariances,centroids       
    return dendrite


#!/usr/bin/env python
# coding: utf-8

# In[4]:


from keras.engine.topology import Layer
from keras import backend as K
from keras import initializers
from keras.initializers import RandomUniform
from keras import activations
import numpy as np
import tensorflow as tf

class DENlayer(Layer):
    
    def __init__(self,ellipses=0, dendrites = [] , activation = None, *args , **kwargs):
        
        self.ellipses = ellipses
        self.dendrites = dendrites
        self.activation = activations.get(activation)
        
        if len(dendrites) == 2:
            self.ellipses =int(np.shape(self.dendrites[0])[0])     
                   
        
        super(DENlayer, self).__init__(**kwargs)
       

            
    def build(self, input_shape):
      
        
        if len(self.dendrites) == 0:
           
            self.Centroids = self.add_weight(name='Centroids', 
                                         shape=(self.ellipses, 1, input_shape[1]), 
                                         initializer = RandomUniform(minval=-0.5, maxval=0.5, seed=None),
                                         trainable = True)
        
            self.Sigmas = self.add_weight(name = 'Sigmas', 
                                          shape = (self.ellipses, input_shape[1], input_shape[1]),
                                          initializer = RandomUniform(minval=0.01, maxval=1, seed=None),
                                          trainable = True)
            
            
        if len(self.dendrites) == 2:
           
            
            self.Centroids = self.add_weight(name='Centroids', 
                                  shape=([self.ellipses, int(np.shape(self.dendrites[1])[1]), int(np.shape(self.dendrites[1])[2])]), 
                                  initializer = initializers.Constant(value = self.dendrites[1]),
                                  trainable = True)
        
            self.Sigmas = self.add_weight(name = 'Sigmas', 
                                  shape = ([self.ellipses, int(np.shape(self.dendrites[0])[1]),
                                            int(np.shape(self.dendrites[0])[2])]),
                                            initializer = initializers.Constant(value = self.dendrites[0]),
                                            trainable = True)
           
           
        
        super(DENlayer, self).build(input_shape) 

    
    def call(self, x):
        
        x = K.expand_dims(x, axis = 0)
        
        x = K.repeat_elements(x, int(np.shape(self.Centroids)[0]), 0)
        
        
        
        dif = x - self.Centroids
        
        difT = K.permute_dimensions(dif, (0, 2, 1))
                
        mah = K.batch_dot(dif, tf.linalg.inv(self.Sigmas))
        mah = K.batch_dot(mah, difT)

        
        diag = tf.linalg.diag_part(mah)
        output = K.permute_dimensions(diag, (1, 0))

       
        if self.activation is not None:
            
           output = self.activation(output)

        return output

    
    def compute_output_shape(self, input_shape):
        
        return (input_shape[0], int(np.shape(self.Centroids)[0]))
      
    def get_config(self):
        
        config = super().get_config().copy()
        config.update({
            'ellipses' : self.ellipses,
            'activation' : self.activation,
        })
        return config
    
    def printWeights(self):
        print("centroids:",self.Centroids)
        print("sigmas:",self.Sigmas)
         

  
    











import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import euclidean_distances


def bkmeans(x_train, y_train, boxes, eps=0.05):

    classes = np.unique(y_train)
    counter = 0
    flag = 0
    

    if len(classes)<=2:
        # Analyzed patterns
        pos = np.where(classes[1] == y_train)[0]
        x = x_train[pos, :]

        # Initial Number of Clusters
        c_num = boxes[counter]

        # Appliying KMeans to the selected class
        kmeans = KMeans(init = 'k-means++', n_clusters = c_num, n_init = 10, max_iter = 100)
        kmeans.fit(x)
        pattern_labels = kmeans.labels_ 
        clusters = np.unique(pattern_labels)
        print("Clusters n",clusters)
                    
        for cluster in clusters:
           
            indices = np.where(cluster == pattern_labels)[0]
            cluster_points = x[indices,:]
            print("clusters",cluster_points)
            print("Dimensiones",len(cluster_points[0]))
            
            
            for d in range(len(cluster_points[0])):
            
                    cluster_d = cluster_points[:,d:d+1]
                   
                    if flag == 0:
                        Wmin_arr = min(cluster_d)-eps
                        Wmax_arr = max(cluster_d)+eps

                    else:
                        Wmin_arr = np.concatenate((Wmin_arr, min(cluster_d)-eps), axis = 0)
                        Wmax_arr = np.concatenate((Wmax_arr, max(cluster_d)+eps), axis = 0)
                    flag += 1 
                        
#        
    
    if len(classes)>2:
        for classe in classes:

            # Analyzed patterns
            pos = np.where(classe == y_train)[0]
            x = x_train[pos, :]

            # Initial Number of Clusters
            c_num = boxes[counter]

            # Appliying KMeans to the selected class
            kmeans = KMeans(init = 'k-means++', n_clusters = c_num, n_init = 10, max_iter = 100)
            kmeans.fit(x)
            pattern_labels = kmeans.labels_ 
            clusters = np.unique(pattern_labels)
            
            
            for cluster in clusters:
               

                indices = np.where(cluster == pattern_labels)[0]
                cluster_points = x[indices,:]
                
                
                #Para cada dimension de cada cluster
                for d in range(len(cluster_points[0])):
                    cluster_d = cluster_points[:,d:d+1]
              
                    if flag == 0:
                        Wmin_arr = min(cluster_d)-eps
                        Wmax_arr = max(cluster_d)+eps

                    else:
                        Wmin_arr = np.concatenate((Wmin_arr, min(cluster_d)-eps), axis = 0)
                        Wmax_arr = np.concatenate((Wmax_arr, max(cluster_d)+eps), axis = 0)
                    flag += 1 

            counter += 1
           

    dendrite = Wmin_arr, Wmax_arr         
    return dendrite




#!/usr/bin/env python
# coding: utf-8

# In[2]:


from keras.engine.topology import Layer
from keras import backend as K
from keras.initializers import Constant
from keras.initializers import RandomUniform
from keras import activations
import numpy as np

class DMNlayer(Layer):
    def __init__(self, units,dendrites = [], activation=None, **kwargs):
        
        
        self.units = units   #Number of dendrites
        self.dendrites = dendrites
        self.activation = activations.get(activation)
        if len(dendrites) == 2:
            self.units = int(len(dendrites[0]))
    
        super(DMNlayer, self).__init__(**kwargs)

    def build(self, input_shape):
        
        
        if len(self.dendrites) == 0:
            
            self.Wmin = self.add_weight(name='Wmin', 
                                      shape=(self.units, input_shape[1]),
                                      initializer=RandomUniform(minval=-2, maxval=-0.1, seed=None),
                                      trainable=True)
            self.Wmax = self.add_weight(name='Wmax', 
                                      shape=(self.units, input_shape[1]),
                                      initializer=RandomUniform(minval=0.1, maxval=0.5, seed=None),
                                      trainable=True)

        if len(self.dendrites) == 2:
            
            self.Wmin = self.add_weight(name='Wmin', 
                                      shape=(int(self.units/input_shape[1]), input_shape[1]),
                                      initializer = Constant(value = self.dendrites[0]),
                                      trainable=True)
           
            self.Wmax = self.add_weight(name='Wmax', 
                                      shape=(int(self.units/input_shape[1]), input_shape[1]),
                                      initializer= Constant(value = self.dendrites[1]),
                                      trainable=True)
        
        
            
            
        super(DMNlayer, self).build(input_shape) 

    def call(self, x):
        
        Q = K.int_shape(x)[0]
        if Q is None: Q = 1
        
      
       
        X = K.repeat(x,np.shape(self.Wmin)[0])
        
        Wmin = K.permute_dimensions(K.repeat(self.Wmin, Q), (1,0,2))
       
        L1 = K.min(X - Wmin, axis=2)
       
        Wmax = K.permute_dimensions(K.repeat(self.Wmax, Q), (1,0,2))
        L2 = K.min(Wmax - X, axis=2)
        output = K.minimum(L1,L2)
        if self.activation is not None:
            output = self.activation(output)
        return output
        
        
    
    def compute_output_shape(self, input_shape):
        return (input_shape[0],self.units) 

    def get_config(self):
        config = super(DMNlayer, self).get_config()
        config.update({"units": self.units,
                       "activation":self.activation})
        return config
    
    def valWmin(self):
        print("Wmin:",self.Wmin)
    
    def valWmax(self):
        print("Wmax:",self.Wmax)





#!/usr/bin/env python
# coding: utf-8

# In[1]:
import numpy as np


def generateWB(P,M):
    if(len(P)==1):
        print("Can't genarete an Hyperbox with one pattern")
    elif(len(P)==0):
        print("Can't genarete an Hyperbox with no patterns")
       
    Wmin = []
    Wmax = []
    
    for i in range(len(P[0])):
        Pclass_x = P[:,i] 
        
        Wmin.append(min(Pclass_x.tolist())-M)
        
        Wmax.append(max(Pclass_x.tolist())+M)   
    return Wmin,Wmax

def HBpC(P,T,M=0):
    classes = np.unique(T)
    
    dendrite = []
    counter = 0
    
    if (len(classes)<=2 and len(classes)>0):
        print("Estamos dentro")
        print(classes)
        for c in range(1):
            indexClasses = np.where(T==c)[0]
            
            if(len(indexClasses)==1):
                continue
            Pclass = P[indexClasses]
                      
            Wmin,Wmax = generateWB(Pclass,M)
            print("Wmin:",Wmin,"Wmax:",Wmax)
            
            if counter == 0:
                Wmin_arr = Wmin
                Wmax_arr = Wmax

#             else:
#                 Wmin_arr = np.concatenate((Wmin_arr, Wmin), axis = 0)
#                 Wmax_arr = np.concatenate((Wmax_arr, Wmax), axis = 0)

#             counter += 1
    
    
    if (len(classes)>2):
        for c in classes:
            indexClasses = np.where(T==c)[0]
            
            if(len(indexClasses)==1):
                continue
            Pclass = P[indexClasses]
                      
            Wmin,Wmax = generateWB(Pclass,M)
            print("Wmin:",Wmin,"Wmax:",Wmax)
            
            if counter == 0:
                Wmin_arr = Wmin
                Wmax_arr = Wmax

            else:
                Wmin_arr = np.concatenate((Wmin_arr, Wmin), axis = 0)
                Wmax_arr = np.concatenate((Wmax_arr, Wmax), axis = 0)

            counter += 1
            
    dendrite = Wmin_arr,Wmax_arr       
    return dendrite




# def generateWB(P,M):
#     if(len(P)==1):
#         print("Can't genarete an Hyperbox with one pattern")
#     elif(len(P)==0):
#         print("Can't genarete an Hyperbox with no patterns")
       
#     Wmin = []
#     Wmax = []
    
#     for i in range(len(P[0])):
#         Pclass_x = P[:,i] 
        
#         Wmin.append(min(Pclass_x.tolist())-M)
        
#         Wmax.append(max(Pclass_x.tolist())+M)   
#     return Wmin,Wmax

# def HBpC(P,T,M=0):
#     classes = np.unique(T)
    
#     dendrite = []
#     counter = 0
    
#     if (len(classes)<=2 and len(classes)>0):
      
#         for c in range(1,2):
#             indexClasses = np.where(T==c)[0]
            
#             if(len(indexClasses)==1):
#                 continue
#             Pclass = P[indexClasses]
                      
#             Wmin,Wmax = generateWB(Pclass,M)
           
            
#             if counter == 0:
#                 Wmin_arr = Wmin
#                 Wmax_arr = Wmax

    
    
#     if (len(classes)>2):
#         for c in classes:
#             indexClasses = np.where(T==c)[0]
            
#             if(len(indexClasses)==1):
#                 continue
#             Pclass = P[indexClasses]
                      
#             Wmin,Wmax = generateWB(Pclass,M)
            
            
#             if counter == 0:
#                 Wmin_arr = Wmin
#                 Wmax_arr = Wmax

#             else:
#                 Wmin_arr = np.concatenate((Wmin_arr, Wmin), axis = 0)
#                 Wmax_arr = np.concatenate((Wmax_arr, Wmax), axis = 0)

#             counter += 1
            
#     dendrite = Wmin_arr,Wmax_arr       
#     return dendrite




#!/usr/bin/env python
# coding: utf-8

# In[1]:
import numpy as np
import math


def generateCR(P,M):
    if(len(P)==1):
        print("Can't genarete an Hypersphere with one pattern")
    elif(len(P)==0):
        print("Can't genarete an Hypersphere with no patterns")
       
    center = []
     
    aux_point = [0]*len(P[0])
    
  
    for i in range(len(P[0])):
        Pclass_x = P[:,i] 
        vmax = max(Pclass_x.tolist())
        vmin = min(Pclass_x.tolist())
        dst = (vmax - vmin) / 2.0
        centroid = vmin + dst
        aux_point[i] = vmax
        center.append(centroid)
       
    suma = 0
    for j in range(len(P[0])):
        suma = suma + (aux_point[j] - center[j])**2
    radius =  math.sqrt( suma   )+M
    return center,radius


def HSpC(P,T,M=0):
    dendrite = []
    counter = 0
    classes = np.unique(T)
    
    if (len(classes)>0 and len(classes)<=2):
        for c in classes:
            indexClasses = np.where(T==c)[0]
            
            if(len(indexClasses)==1):
                continue
            Pclass = P[indexClasses]
                      
            center,rad = generateCR(Pclass,M)
            
           
            if counter == 0:
                center_arr = center
                rad_arr = [rad]

            
    if (len(classes)>2):
        for c in classes:
            indexClasses = np.where(T==c)[0]
            
            if(len(indexClasses)==1):
                continue
            Pclass = P[indexClasses]
                      
            center,rad = generateCR(Pclass,M)
          
           
            if counter == 0:
                center_arr = center
                rad_arr = [rad]

            else:
                center_arr = np.concatenate((center_arr, center), axis = 0)
                rad_arr.append(rad)

            counter += 1
            
    dendrite = center_arr,rad_arr       
    return dendrite


#!/usr/bin/env python
# coding: utf-8

# In[6]:


from keras.engine.topology import Layer
from keras import activations
from keras.initializers import RandomUniform
from keras import backend as K
import numpy as np
import tensorflow as tf
from keras import initializers

class DSNlayer(Layer):
    
    def __init__(self, spheres, dendrites = [], activation = None, **kwargs):
        
        self.spheres = spheres
        self.dendrites = dendrites
        self.activation = activations.get(activation)
        if len(dendrites) == 2:
            self.spheres = len(dendrites[1])
        
        
        
        super(DSNlayer, self).__init__(**kwargs)

        
    def build(self, input_shape):
        
        
        if len(self.dendrites) == 0:
           
            
            self.centroid = self.add_weight(name='Centroids', 
                                      shape=(self.spheres,1,input_shape[1]), 
                                      initializer = RandomUniform(minval=-1, maxval=1, seed=None),
                                      trainable = True)
           
        
            self.radius = self.add_weight(name = 'Radius', 
                                      shape = (self.spheres,1),
                                      initializer = RandomUniform(minval=-1, maxval=1, seed=None),
                                      trainable = True)
           
            
        if len(self.dendrites) == 2:
            
                
            self.centroid = self.add_weight(name='Centroids', 
                                      shape=(self.spheres,1,input_shape[1]), 
                                      initializer = initializers.Constant(value = self.dendrites[0]),
                                      trainable = True)
            
        
            self.radius = self.add_weight(name = 'Radius', 
                                      shape = (self.spheres,1),
                                      initializer = initializers.Constant(value = self.dendrites[1]),
                                      trainable = True)
                
        
        super(DSNlayer, self).build(input_shape) 

    
    def call(self, x):
        
        x = K.expand_dims(x, axis = 0)
       
        x = K.repeat_elements(x,int(np.shape(self.centroid)[0]),0)
      
        dif = x-self.centroid
       
        distance = tf.norm(dif,axis=2)
      
        
        output=self.radius-(distance)
        output = K.permute_dimensions(output, (1, 0))
        
        if self.activation is not None:
          output = self.activation(output)
        return output
    
    def compute_output_shape(self, input_shape):
        
        return (input_shape[0], int(np.shape(self.radius)[0]))
      
    def get_config(self):
        
        config = super().get_config().copy()
        config.update({
            'spheres' : self.spheres,
            'activation' : self.activation,
        })
        return config


# In[ ]:







from setuptools import setup
from setuptools import find_packages

setup(
    name = 'Dendrite Neural Network',
    version = '0.0.1',
    description = 'Probando...',
    py_modules = ["DMN","DEN","DSN","PreTrain.HpC.HBpC","PreTrain.HpC.HEpC","PreTrain.HpC.HSpC", "PreTrain.kmeans.bkmeans","PreTrain.kmeans.ekmeans","PreTrain.kmeans.skmeans"],
    packages = find_packages(),
    package_dir = {'':'src'}

)



import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import euclidean_distances

def ekmeans(x_train, y_train, ellipses):

    classes = np.unique(y_train)
    counter = 0
    
    if len(classes)<=2:
        # Analyzed patterns
        pos = np.where(classes[0] == y_train)[0]
        x = x_train[pos, :]

        # Initial Number of Clusters
        c_num = ellipses[counter]

        # Appliying KMeans to the selected class
        kmeans = KMeans(init = 'k-means++', n_clusters = c_num, n_init = 10, max_iter = 100)
        kmeans.fit(x)
        centroids = np.reshape(kmeans.cluster_centers_, (c_num, 1, np.shape(x_train)[1]))
        print("centroid:",centroids)
        pattern_labels = kmeans.labels_ 
        clusters = np.unique(pattern_labels)
        covariances = np.zeros((c_num, np.shape(x_train)[1], np.shape(x_train)[1]))
        
        for cluster in clusters:

            indices = np.where(cluster == pattern_labels)[0]
            covariances[cluster] = np.cov(x[indices, :].T)

        print("Covariances:",covariances)

        if counter == 0:
            covariances_arr = covariances
            centroids_arr = centroids

        else:
            covariances_arr = np.concatenate((covariances_arr, covariances), axis = 0)
            centroids_arr = np.concatenate((centroids_arr, centroids), axis = 0)

        counter += 1

    
    if len(classes)>2:
        for classe in classes:

            # Analyzed patterns
            pos = np.where(classe == y_train)[0]
            x = x_train[pos, :]

            # Initial Number of Clusters
            c_num = ellipses[counter]

            # Appliying KMeans to the selected class
            kmeans = KMeans(init = 'k-means++', n_clusters = c_num, n_init = 10, max_iter = 100)
            kmeans.fit(x)
            centroids = np.reshape(kmeans.cluster_centers_, (c_num, 1, np.shape(x_train)[1]))
           
            pattern_labels = kmeans.labels_ 
            clusters = np.unique(pattern_labels)
            covariances = np.zeros((c_num, np.shape(x_train)[1], np.shape(x_train)[1]))
            
            for cluster in clusters:

                indices = np.where(cluster == pattern_labels)[0]
                covariances[cluster] = np.cov(x[indices, :].T,bias=True)

            

            if counter == 0:
                covariances_arr = covariances
                centroids_arr = centroids

            else:
                covariances_arr = np.concatenate((covariances_arr, covariances), axis = 0)
                centroids_arr = np.concatenate((centroids_arr, centroids), axis = 0)

            counter += 1
            
    dendrite = covariances_arr, centroids_arr
    return dendrite

